# Copyright (c) 2021-2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import threading
import time
import math


def nop_callback(_sweep, _pos, _sd_results):
    pass


class SleepInterruptedException(Exception):
    pass


class Sweeper:
    def __init__(self, tc, amplitude, f0, f1, nfreqs, sweep_time,
                 min_freq_time, log10=True, settle_ms=2,
                 data_callback=nop_callback, nsweeps=0):
        self.sleep_cond     = threading.Condition()
        self.tc             = tc
        self.amplitude      = amplitude
        self.f0             = f0
        self.f1             = f1
        self.min_f          = min(f0, f1)
        self.max_f          = max(f0, f1)
        self.nfreqs         = nfreqs
        self.log10          = log10
        self.settle_ms      = settle_ms
        self.data_callback  = data_callback
        self.nsweeps        = nsweeps
        self.running        = False

        N = self.nfreqs
        if self.log10:
            k     = self.f1 / self.f0
            freqs = [self.f0 * k**(i/(N-1)) for i in range(N)]
        else:
            df    = (self.f1 - self.f0) / N
            freqs = [self.f0 + i*df for i in range(N)]

        fixed_time   = min_freq_time * N
        dynamic_time = sweep_time - fixed_time
        p = dynamic_time / sum(1 / f for f in freqs)
        self.freq_tuples = [(f, math.ceil(1000 * (min_freq_time + p / f)))
                            for f in freqs]

    def start(self):
        with self.sleep_cond:
            self.running = True
            threading.Thread(target=self.usb_thread).start()

    def stop(self):
        with self.sleep_cond:
            self.running = False
            self.sleep_cond.notify()

    def sleep_interruptible(self, dt, verbose=True):
        if verbose:
            wakeup_time = time.localtime(time.time() + dt)
            self.tc.info('Sleeping for %.3f seconds until %s...'
                         % (dt, time.asctime(wakeup_time)))

        with self.sleep_cond:
            start_time = time.time()
            while True:
                now     = time.time()
                elapsed = now - start_time
                if elapsed >= dt:
                    break
                if not self.running:
                    raise SleepInterruptedException()

                rem = dt - elapsed
                self.sleep_cond.wait(timeout=rem)

    def usb_thread(self):
        self.tc.info('Sweeping %u frequencies from %s Hz to %s Hz...'
                     % (self.nfreqs, self.f0, self.f1))

        pos   = 0
        sweep = 1
        while self.running:
            block  = self.freq_tuples[pos:pos+self.tc.ginfo.max_sweep_entries]
            nfreqs = len(block)
            t0_ns  = time.time_ns()
            dt, _  = self.tc.sweep_async(self.amplitude, block,
                                         ndiscards=self.settle_ms)
            try:
                self.sleep_interruptible(dt)
            except SleepInterruptedException:
                break
            sd = self.tc.read_sweep_data()

            for r, b in zip(sd.results, block):
                assert r.nbufs == b[1]

            self.data_callback(sweep, t0_ns, pos, sd.results, self.amplitude)

            if pos + nfreqs == len(self.freq_tuples):
                self.tc.info('Sweep %u complete.' % sweep)
                pos = 0
                if sweep == self.nsweeps:
                    break
                sweep += 1
            else:
                pos += nfreqs

        self.tc.info('Sweeper stopped.')
        self.running = False



# Copyright (c) 2022-2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import xtalx.tools.usb

from .tcsc_u5 import TCSC_U5
from .peak_tracker import PeakTracker
from .predicate_queue import PredicateQueue
from .sweeper import Sweeper


def find(**kwargs):
    return xtalx.tools.usb.find(idVendor=0x0483, idProduct=0xA34E,
                                product='XtalX TCSC', find_all=True, **kwargs)


def find_one(**kwargs):
    return xtalx.tools.usb.find_one(idVendor=0x0483, idProduct=0xA34E,
                                    product='XtalX TCSC', find_all=True,
                                    **kwargs)


def make(usb_dev, **kwargs):
    if usb_dev.product == 'XtalX TCSC':
        return TCSC_U5(usb_dev, **kwargs)

    raise Exception('Unrecognized product string: %s' % usb_dev.product)


__all__ = ['find',
           'find_one',
           'make',
           'PeakTracker',
           'PredicateQueue',
           'Sweeper',
           ]



# Copyright (c) 2021-2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import time


class TinCan:
    def __init__(self, verbose=False, yield_Y=False):
        self.verbose         = verbose
        self.yield_Y         = yield_Y

    def log(self, tag, s, timestamp=None):
        timestamp = timestamp or time.time_ns()
        print('[%u] %s: %s' % (timestamp, tag, s))

    def info(self, s, **kwargs):
        self.log('I', s, **kwargs)

    def warn(self, s, **kwargs):
        self.log('W', s, **kwargs)

    def vlog(self, tag, s, **kwargs):
        if self.verbose:
            self.log(tag, s, **kwargs)

    def dbg(self, s, **kwargs):
        self.vlog('D', s, **kwargs)



# Copyright (c) 2020-2023 by Phase Advanced Sensor Systems Corp.
import threading
import errno
import usb
import usb.util

import btype

from .exception import XtalXException


FC_FLAGS_VALID              = (1 << 15)
FC_FLAG_NO_TEMP_PRESSURE    = (1 << 4)
FC_FLAG_PRESSURE_FAILED     = (1 << 3)
FC_FLAG_TEMP_FAILED         = (1 << 2)
FC_FLAG_PRESSURE_UPDATE     = (1 << 1)
FC_FLAG_TEMP_UPDATE         = (1 << 0)


class FrequencyPacket24(btype.Struct):
    '''
    Firmware revisions 1.0.6 and earlier return a 24-byte packet if the sensor
    doesn't have enough data to perform a temperature-compensated pressure
    measurement yet or if the sensor doesn't have a calibration applied in
    flash.
    '''
    ref_freq            = btype.uint32_t()
    pressure_edges      = btype.uint32_t()
    pressure_ref_clocks = btype.uint32_t()
    temp_edges          = btype.uint32_t()
    temp_ref_clocks     = btype.uint32_t()
    flags               = btype.uint16_t()
    seq_num             = btype.uint8_t()
    rsrv                = btype.uint8_t()
    _EXPECTED_SIZE      = 24


class FrequencyPacket40(btype.Struct):
    '''
    Firmware revisions 1.0.6 and earlier return a 40-byte packet if the sensor
    has enough data to perform a temperature-compensated pressure measurement.
    '''
    ref_freq            = btype.uint32_t()
    pressure_edges      = btype.uint32_t()
    pressure_ref_clocks = btype.uint32_t()
    temp_edges          = btype.uint32_t()
    temp_ref_clocks     = btype.uint32_t()
    flags               = btype.uint16_t()
    seq_num             = btype.uint8_t()
    rsrv                = btype.uint8_t()
    pressure_psi        = btype.float64_t()
    temp_c              = btype.float64_t()
    _EXPECTED_SIZE      = 40


class FrequencyPacket56(btype.Struct):
    '''
    Firmware revisions 1.0.7 and higher always return a 56-byte packet that
    contains flags indicating the validity of things like the temperature-
    compensated pressure measurement.  These firmware versions also return the
    MCU temperature as a control.
    '''
    ref_freq            = btype.uint32_t()
    pressure_edges      = btype.uint32_t()
    pressure_ref_clocks = btype.uint32_t()
    temp_edges          = btype.uint32_t()
    temp_ref_clocks     = btype.uint32_t()
    flags               = btype.uint16_t()
    seq_num             = btype.uint8_t()
    rsrv                = btype.uint8_t()
    pressure_psi        = btype.float64_t()
    temp_c              = btype.float64_t()
    mcu_temp_c          = btype.float64_t()
    rsrv2               = btype.Array(btype.uint8_t(), 8)
    _EXPECTED_SIZE      = 56


class Measurement:
    '''
    Object encapsulating the results of an XTI sensor measurement.  The
    following fields are defined:

        sensor - Reference to the XTI that generated the Measurement.
        ref_freq - Frequency of the sensor's reference crystal.
        pressure_edges - Number of pressure crystal ticks used to generate the
            Measurement.
        pressure_ref_clocks - Number of reference clock ticks that elapsed
            while counting pressure_edges pressure crystal ticks.
        pressure_freq - Measured pressure crystal frequency.
        temp_edges - Number of temperature crystal ticks used to generate the
            Measurement.
        temp_ref_clocks - Number of temperature crystal ticks that elapsed
            while counting temp_edges temperature crystal ticks.
        temp_freq - Measured temperature crystal frequency.
        mcu_temp_c - Microcontroller's internal junction temperature.
        pressure_psi - Temperature-compensated pressure measured in PSI.
        temp_c - Temperature measured in degrees Celsius.
        flags - A set of validity and error flags.

    If the sensor is uncalibrated or has not sampled enough data to generate
    a temperature-compensated pressure measurement then some or all of
    temp_freq, pressure_freq, pressure_psi and temp_c may be None.

    The flags field is a bitmask which may include any of the following bits;
    it may be None if the firmware version predates the introduction of status
    flags:

        FC_FLAGS_VALID - The flags field contains valid information (always set
            or flags will be None).
        FC_FLAG_NO_TEMP_PRESSURE - Will be set if pressure_psi and temp_c could
            not be generated; the sensor may be uncalibrated or may not have
            generated both temperature and pressure crystal readings yet.
        FC_FLAG_PRESSURE_FAILED - Will be set if 0.5 seconds elapse without a
            pressure crystal measurement completing; this indicates that a
            sensor failure has caused the pressure crystal to stop ticking.
        FC_FLAG_TEMP_FAILED - Will be set if 0.5 seconds elapse without a
            temperature crystal measurement completing; this indicates that a
            sensor failure has caused the temperature crystal to stop ticking.
        FC_FLAG_PRESSURE_UPDATE - Indicates that the current Measurement
            incorporates a new reading from the pressure crystal; it may still
            be incorporating the previous reading from the temperature crystal.
        FC_FLAG_TEMP_UPDATE - Indicates that the current Measurement
            incorporates a new reading from the temperature crystal; it may
            still be incorporating the previous reading from the pressure
            crystal.

    Note that since the temperature and pressure crystals tick asynchronously
    with respect to one another, a measurement on one crystal is likely to
    complete while a measurement on the other crystal is still pending and so
    typically only one of FC_FLAG_PRESSURE_UPDATE or FC_FLAG_TEMP_UPDATE will
    be set.
    '''
    def __init__(self, sensor, ref_freq, pressure_edges, pressure_ref_clocks,
                 temp_edges, temp_ref_clocks, mcu_temp_c, pressure_psi,
                 temp_c, flags):
        self.sensor              = sensor
        self.ref_freq            = ref_freq
        self.pressure_edges      = pressure_edges
        self.pressure_ref_clocks = pressure_ref_clocks
        self.temp_edges          = temp_edges
        self.temp_ref_clocks     = temp_ref_clocks
        self.mcu_temp_c          = mcu_temp_c
        self.pressure_psi        = pressure_psi
        self.temp_c              = temp_c
        self.flags               = flags

        if temp_ref_clocks > 3:
            self.temp_freq = ref_freq * temp_edges / temp_ref_clocks
        else:
            self.temp_freq = None

        if pressure_ref_clocks > 3:
            self.pressure_freq = ref_freq * pressure_edges / pressure_ref_clocks
        else:
            self.pressure_freq = None

    @staticmethod
    def _from_packet(sensor, packet):
        mt, p, t = None, None, None
        if sensor.usb_dev.bcdDevice < 0x0107:
            if len(packet) == 24:
                fp = FrequencyPacket24.unpack(packet)
            else:
                fp = FrequencyPacket40.unpack(packet)
                p  = fp.pressure_psi
                t  = fp.temp_c
        else:
            fp = FrequencyPacket56.unpack(packet)
            mt = fp.mcu_temp_c
            assert fp.flags and (fp.flags & FC_FLAGS_VALID)
            if (fp.flags & FC_FLAG_NO_TEMP_PRESSURE) == 0:
                p = fp.pressure_psi
                t = fp.temp_c
        flags = fp.flags if fp.flags & FC_FLAGS_VALID else None

        return Measurement(sensor, fp.ref_freq, fp.pressure_edges,
                           fp.pressure_ref_clocks, fp.temp_edges,
                           fp.temp_ref_clocks, mt, p, t, flags)

    def tostring(self, verbose=False):
        s = '%s: ' % self.sensor
        if verbose:
            s += ('C %u pe %u prc %u pf %f te %u trc %u tf %f p %s t %s '
                  'mt %s' % (self.ref_freq, self.pressure_edges,
                             self.pressure_ref_clocks, self.pressure_freq,
                             self.temp_edges, self.temp_ref_clocks,
                             self.temp_freq, self.pressure_psi, self.temp_c,
                             self.mcu_temp_c))
        else:
            if self.pressure_psi is None:
                p = 'n/a'
            else:
                p = '%f' % self.pressure_psi

            if self.temp_c is None:
                t = 'n/a'
            else:
                t = '%f' % self.temp_c
            s += '%s PSI, %s C' % (p, t)

        return s


class XTI:
    '''
    Given a USB device handle acquired via find() or find_one(), creates an
    XTI object that can be used to communicate with a sensor.
    '''
    def __init__(self, usb_dev):
        self.usb_dev     = usb_dev
        self.lock        = threading.RLock()
        self._halt_yield = True
        self.thread      = None

        try:
            self.serial_num = usb_dev.serial_number
            self.git_sha1   = usb.util.get_string(usb_dev, 6)
            self.fw_version = usb_dev.bcdDevice
        except ValueError as e:
            if str(e) == 'The device has no langid':
                raise XtalXException(
                    'Device has no langid, ensure running as root!') from e

        if self.usb_dev.bcdDevice >= 0x0103:
            try:
                self.report_id = usb.util.get_string(usb_dev, 15)
            except ValueError:
                self.report_id = None
        else:
            self.report_id = None

        self.usb_path = '%s:%s' % (
            usb_dev.bus, '.'.join('%u' % n for n in usb_dev.port_numbers))

    def __str__(self):
        return 'XTI(%s)' % self.serial_num

    def _set_configuration(self, bConfigurationValue):
        with self.lock:
            cfg = None
            try:
                cfg = self.usb_dev.get_active_configuration()
            except usb.core.USBError as e:
                if e.strerror != 'Configuration not set':
                    raise

            if cfg is None or cfg.bConfigurationValue != bConfigurationValue:
                usb.util.dispose_resources(self.usb_dev)
                self.usb_dev.set_configuration(bConfigurationValue)

    def _set_measurement_config(self):
        self._set_configuration(2)

    def read_measurement(self):
        '''
        Synchronously read a single measurement from the sensor, blocking if no
        measurement is currently available.
        '''
        with self.lock:
            p = self.usb_dev.read(0x81, 64)
        return Measurement._from_packet(self, p)

    def _yield_measurements(self, do_reset):
        with self.lock:
            if do_reset:
                self.usb_dev.reset()
            self._set_measurement_config()

            while not self._halt_yield:
                try:
                    yield self.read_measurement()
                except usb.core.USBError as e:
                    if e.errno != errno.ETIMEDOUT:
                        raise
                    continue

    def yield_measurements(self, do_reset=True):
        '''
        Yields Measurement objects synchronously in the current thread,
        blocking while waiting for new measurements to be acquired.
        '''
        with self.lock:
            self._halt_yield = False
            yield from self._yield_measurements(do_reset)

    def halt_yield(self):
        '''
        Halts an ongoing yield_measurements() call, causing it to eventually
        terminate the generator loop.
        '''
        self._halt_yield = True

    def _read_measurements_async(self, handler, do_reset):
        with self.lock:
            for m in self._yield_measurements(do_reset):
                handler(m)

    def read_measurements(self, handler, do_reset=True):
        '''
        Reads measurements asynchronously in a separate thread, calling the
        handler as measurements become available.  The handler should take a
        single Measurement object as an argument.
        '''
        with self.lock:
            assert self.thread is None
            self._halt_yield = False
            self.thread = threading.Thread(target=self._read_measurements_async,
                                           args=(handler, do_reset),
                                           daemon=False)
            self.thread.start()

    def join_read(self):
        '''
        Blocks the current thread until the asynchronous read thread completes.
        Typically this blocks indefinitely until some error occurs, however the
        read thread will also exit if someone sets the _halt_yield field to
        True (see XTI.halt_read()).
        '''
        self.thread.join()

    def halt_read(self):
        '''
        Halts any asynchronous measurement thread and waits for it to finish
        cleanly.
        '''
        self._halt_yield = True
        self.join_read()



# Copyright (c) 2020-2023 by Phase Advanced Sensor Systems Corp.
import xtalx.tools.usb

from .xti import XTI


def find_xti(**kwargs):
    return xtalx.tools.usb.find(idVendor=0x0483, idProduct=0xA34E,
                                product='XtalX', find_all=True, **kwargs)


def find_one_xti(**kwargs):
    return xtalx.tools.usb.find_one(idVendor=0x0483, idProduct=0xA34E,
                                    product='XtalX', find_all=True, **kwargs)


def make(usb_dev, **kwargs):
    if usb_dev.product == 'XtalX':
        return XTI(usb_dev, **kwargs)

    raise Exception('Unrecognized product string: %s' % usb_dev.product)


__all__ = ['find_xti',
           'find_one_xti',
           'make',
           ]



#!/usr/bin/env python3
# Copyright (c) 2021-2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import argparse
import threading
import cmath
import math

import glotlib
import numpy as np

import xtalx.z_sensor


class SweepWindow(glotlib.Window):
    def __init__(self, tc, *args, dump_fpath=None, **kwargs):
        super().__init__(900, 700, msaa=4, name=tc.serial_num or '')

        self.tc          = tc
        self.data_lock   = threading.Lock()
        self.data        = []
        self.data_gen    = -1
        self.plot_gen    = -1
        self.num_samples = 0
        self.snap        = True

        self.sweeper = xtalx.z_sensor.Sweeper(tc, *args,
                                              data_callback=self.data_callback,
                                              **kwargs)

        if dump_fpath:
            print('Dumping to %s' % dump_fpath)
            self.file = open(  # pylint: disable=R1732
                    dump_fpath, 'a', encoding='utf8')
            self.file.write('sweep,time_ns,f,dt,zx_real,zx_imag,RR,amplitude\n')
            self.file.flush()
        else:
            self.file = None

        # |Zx| plot.
        self.zx_plot = self.add_plot(
            321, limits=(self.sweeper.min_f, 0, self.sweeper.max_f, 1))
        self.zx_lines = self.zx_plot.add_lines([])
        self.zx_plot.add_hline(0, color=(0.5, 0.75, 0.5))

        # Phase plot.
        self.phi_plot = self.add_plot(
            323, limits=(self.sweeper.min_f, -math.pi,
                         self.sweeper.max_f, math.pi), sharex=self.zx_plot)
        self.phi_lines = self.phi_plot.add_lines([])
        self.phi_plot.add_hline(0, color=(0.5, 0.75, 0.5))

        # R**2 plot.
        self.rr_plot = self.add_plot(
            325, limits=(self.sweeper.min_f, 0, self.sweeper.max_f, 1.01),
            sharex=self.zx_plot)
        self.rrs_lines = [self.rr_plot.add_lines([]) for _ in tc.ADC_KEYS]

        # ADCs plot.
        self.adcs_plot = self.add_plot(
            322, limits=(self.sweeper.min_f, 0, self.sweeper.max_f, 65536),
            sharex=self.zx_plot)
        self.adcs_lines = [self.adcs_plot.add_lines([]) for _ in tc.ADC_KEYS]
        self.adcs_plot.add_hline(0, color=(0.5, 0.75, 0.5))

        # Nyquist plot.
        self.nyq_plot = self.add_plot(
            (3, 2, (4, 6)), aspect=glotlib.ASPECT_SQUARE)
        self.nyq_lines = self.nyq_plot.add_lines([], point_width=3)
        self.nyq_plot.add_hline(0, color=(0.5, 0.75, 0.5))
        self.nyq_plot.add_vline(0, color=(0.5, 0.75, 0.5))

    def start(self):
        self.sweeper.start()

    def stop(self):
        self.sweeper.stop()

    def update_geometry(self, _t):
        with self.data_lock:
            if self.data_gen == self.plot_gen:
                return False

            self.plot_gen = self.data_gen
            sdr           = self.data
            snap          = self.snap

        self.tc.info('Updating plot...')

        self.num_samples += 1

        fs = np.array([r.f for r in sdr])
        zrs = np.array([r.z.real for r in sdr])

        for i, rr_lines in enumerate(self.rrs_lines):
            rr_lines.set_x_y_data(fs, [r.RR[i] for r in sdr])

        self.zx_lines.set_x_y_data(fs, [abs(r.z) for r in sdr])
        self.phi_lines.set_x_y_data(fs, [cmath.phase(r.z) for r in sdr])
        self.nyq_lines.set_x_y_data(zrs, [-r.z.imag for r in sdr])

        if snap:
            self.zx_plot.snap_bounds()
            self.nyq_plot.snap_bounds()

        # for i, adc_lines in enumerate(self.adcs_lines):
        #     adc_lines.set_x_y_data(fs, [r.amplitude[i] for r in sdr])
        self.adcs_lines[0].set_x_y_data(fs, [r.z.real for r in sdr])

        return True

    def data_callback(self, sweep, t0_ns, pos, sd_results, amplitude):
        with self.data_lock:
            self.data[pos:pos+len(sd_results)] = sd_results
            self.data_gen                     += 1
            self.snap = (sweep == 1 or (sweep == 2 and pos == 0))

        if self.file:
            for r in sd_results:
                self.file.write('%s,%u,%s,%.3f,%s,%s,%s,%u\n' %
                                (sweep, t0_ns, r.f, r.nbufs / 1000., r.z.real,
                                 r.z.imag, r.RR[1], amplitude))
            self.file.flush()

        self.mark_dirty()


def main(rv):
    min_time = rv.min_freq_time_secs * rv.nfreqs
    if rv.sweep_time_secs < min_time:
        raise Exception('Sweep time must be at least %s seconds for this '
                        'choice of parameters.' % min_time)

    f0 = rv.freq_0
    f1 = rv.freq_1
    track_admittance = not rv.track_impedance

    dev = xtalx.z_sensor.find_one(serial_number=rv.sensor)
    tc  = xtalx.z_sensor.make(dev, verbose=rv.verbose, yield_Y=track_admittance)

    if track_admittance:
        tc.info('Tracking admittance.')
    else:
        tc.info('Tracking impedance.')

    if not rv.amplitude:
        amplitude = tc.cal_dac_amp()
        if amplitude is None:
            raise Exception("Calibration page doesn't include the DAC voltage "
                            "under which the calibration was performed, must "
                            "specify --amplitude manually.")
    elif rv.amplitude.upper().endswith('V'):
        volts = float(rv.amplitude[:-1])
        amplitude = tc.a_to_dac(volts)
        if amplitude is None:
            raise Exception("Calibration page doesn't have required voltage-"
                            "to-DAC information to use amplitudes in Volts.")
        amplitude = round(amplitude)
    else:
        amplitude = int(rv.amplitude)

    tc.info('Using amplitude of %u DAC codes.' % amplitude)
    volts = tc.dac_to_a(amplitude)
    if volts is not None:
        tc.info('Using amplitude of %f V.' % volts)

    if not 0 <= amplitude <= 2000:
        raise Exception('Amplitude not in range 0 to 2000.')

    sw  = SweepWindow(tc, amplitude, f0, f1, rv.nfreqs, rv.sweep_time_secs,
                      rv.min_freq_time_secs, log10=rv.log10,
                      dump_fpath=rv.dump_file, settle_ms=rv.settle_ms,
                      nsweeps=rv.nsweeps)
    sw.start()

    try:
        glotlib.interact()
    except KeyboardInterrupt:
        print()
    finally:
        sw.stop()


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--amplitude', '-a')
    parser.add_argument('--freq-0', '-0', type=float, required=True)
    parser.add_argument('--freq-1', '-1', type=float, required=True)
    parser.add_argument('--nfreqs', '-n', type=int, default=1000)
    parser.add_argument('--nsweeps', type=int, default=0)
    parser.add_argument('--log10', action='store_true')
    parser.add_argument('--verbose', '-v', action='store_true')
    parser.add_argument('--dump-file', '-f')
    parser.add_argument('--sweep-time-secs', type=float, required=True)
    parser.add_argument('--min-freq-time-secs', type=float, default=0)
    parser.add_argument('--settle-ms', type=int, default=2)
    parser.add_argument('--sensor')
    parser.add_argument('--track-impedance', action='store_true')
    rv = parser.parse_args()

    try:
        main(rv)
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()
